//
//  SquishController.mm
//  squish
//
//  Created by Da Woon Jung on 2006-06-23.
//  Copyright 2006-9 Da Woon Jung. All rights reserved.
//

#import <Carbon/Carbon.h>
#import <unistd.h>
#import "SquishController.h"
#import "SquishExportOptionsController.h"
#import "SquishProgressController.h"
#import "SquishCompressNotifier.h"
#import "SquishDDSWindow.h"
#import "NSImage_Extensions.h"
#import "dds.h"
#import "imageutils.h"
#import <squish/squish.h>
using namespace squish;

#define USE_FASTFLOOR

#ifdef USE_FASTFLOOR
// fastfloor() using magic numbers based on Sree Kotay's xs_FloorToInt()
// http://www.stereopsis.com/sree/fpu2006.html
const double _xs_doublemagic         = 6755399441055744.0;
//2^52 * 1.5, uses limited precision to floor
const double _xs_doublemagicdelta    = 1.5e-8;
//almost .5f = .5f + 1e^(number of exp bit)
const double _xs_doublemagicroundeps = .5f-_xs_doublemagicdelta;
//almost .5f = .5f - 1e^(number of exp bit)
#ifdef __BIG_ENDIAN__
#define _xs_iman_                   1
#else
#define _xs_iman_                   0   //intel is little endian
#endif
typedef union
{
    double r64;
    int  i32[2];
} double_or_int;

inline int fastfloor(double x)
{
    double_or_int *u = (double_or_int *)&x;
    u->r64 = x - _xs_doublemagicroundeps + _xs_doublemagic;
    return u->i32[_xs_iman_];
}

#define FLOOR(x)    fastfloor(x)
#else
#define FLOOR(x)    floor(x)
#endif

#ifdef USE_LIBPNG
void PngWrite(png_structp png_ptr, png_bytep data, png_size_t length)
{
}
void PngFlush(png_structp png_ptr)
{
}
#endif

#define DDS_EXT     @"dds"
#define DXT5NM_EXT  @"dxt5nm"
NSString *MIPMAP_EXT[] = {
    @"tiff",
    @"png"
};

enum {
    kDxt5nm = 1 << 3
};

enum {
    kFileModeDDS,
    kFileModeMipmap
};

@interface SquishController (Private)
- (NSApplicationDelegateReply)confirmReplace: (NSString *)aOutFile
                                 refreshTime: (NSDate **)aTime;
- (NSApplicationDelegateReply)exportDDSFromFiles: (NSArray *)files
                                  representation: (NSBitmapImageRep *)rep
                                          toFile: (NSString *)aOutFile
                                          ofType: (NSString *)aType
                                       withFlags: (int)aFlags;
- (NSApplicationDelegateReply)exportMipmapsFromFile: (NSString *)path
                                     representation: (NSBitmapImageRep *)rep
                                             toFile: (NSString *)aOutFile
                                             format: (int)mipmapFormat;
+ (NSBitmapImageRep *)mipmapFromRepresentation: (NSBitmapImageRep *)rep
                                         width: (int)width
                                        height: (int)height;
+ (NSArray *)extensions;
- (int)flagsFromPrefs;
- (NSString *)pathForOutputFile: (NSString *)aFile
                      extension: (NSString *)aExt
                   outputFolder: (NSString *)aDest;
- (NSApplicationDelegateReply)compressBytes: (unsigned char *)aBytes
                                    toBlock: (u8 *)output
                                      width: (size_t)aWidth
                                     height: (size_t)aHeight
                                bytesPerRow: (size_t)aBytesPerRow
                            samplesPerPixel: (int)aSamplesPerPixel
                                      alpha: (int)aAlpha
                                  withFlags: (int)aFlags;
- (NSApplicationDelegateReply)compressBytes: (unsigned char *)aBytes
                                     toFile: (FILE *)ofp
                                      width: (size_t)aWidth
                                     height: (size_t)aHeight
                                bytesPerRow: (size_t)aBytesPerRow
                            samplesPerPixel: (int)aSamplesPerPixel
                                      alpha: (int)aAlpha
                                premultiply:
#ifdef USE_AUTO_PREMULTIPLY
                                             (BOOL)aPremultiply
#else
                                             (int)aPremulOption
#endif
                                  withFlags: (int)aFlags;
- (NSApplicationDelegateReply)writeDDSHeaderToFile: (FILE *)ofp
                                            ofType: (NSString *)aType
                                             width: (size_t)aWidth
                                            height: (size_t)aHeight
                                         withFlags: (int)aFlags
                                           mipmaps: (BOOL)genMipmaps;
- (NSApplicationDelegateReply)decompressBytes: (u8 *)aBytes
                                      toBlock: (u8 **)output
                                        width: (size_t)aWidth
                                       height: (size_t)aHeight
                                    withFlags: (int)aFlags;
@end


@implementation SquishController

- (void)awakeFromNib
{
    CMGetDefaultProfileByUse(cmDisplayUse, &dispProf);
}

- (void)applicationWillFinishLaunching:(NSNotification *)aNotification
{
    [SquishExportOptionsController registerDefaults];
    showOptions = (0 != (GetCurrentKeyModifiers() & optionKey));
}

- (void)applicationDidFinishLaunching:(NSNotification *)aNotification
{
    if (didOpenFiles)
        [[aNotification object] performSelector: @selector(terminate:) withObject: nil afterDelay: 0.1];
    else
        showOptions = YES;
}

- (void)applicationWillTerminate:(NSNotification *)aNotification
{
#ifdef USE_DDS_PREVIEW
    [ddsPreview release];
#endif
    [exportController release];
    CMCloseProfile(dispProf);
}

- (BOOL)applicationShouldOpenUntitledFile:(NSApplication *)aSender
{
    return NO;
}

- (void)application:(NSApplication *)aSender openFiles:(NSArray *)aFilenames
{
    NSApplicationDelegateReply reply = NSApplicationDelegateReplyFailure;
    didOpenFiles = YES;
    [self setEnableGenMipmaps: YES];
    [self setEnableAssembleMipmaps: YES];

    if (aFilenames && [aFilenames count] > 0)
    {
        if ([aFilenames count] < 2)
        {
            [[NSUserDefaults standardUserDefaults] setBool:NO forKey:@"assembleMipmaps"];
            [self setEnableAssembleMipmaps: NO];
        }

        if (showOptions)
        {
            if (NSRunAbortedResponse == [self showExportOptions])
            {
                // Do not reply "NSApplicationDelegateReplyCancel"
                // because doing so causes the filename to disappear
                // from the Recent Items list!
                [aSender replyToOpenOrPrint: NSApplicationDelegateReplySuccess];
                return;
            }
        }

        if (NSApplicationDelegateReplyCancel != reply)
        {
            reply = [self exportFiles: aFilenames];
        }
    }

    [aSender replyToOpenOrPrint: reply];
}


- (void)browseFilesOrFolders: (id)aSender
{
    NSOpenPanel *oPanel = [NSOpenPanel openPanel];
    [oPanel setAllowsMultipleSelection: YES];
    [oPanel setCanChooseDirectories: YES];
    [oPanel setCanChooseFiles: YES];
    
    if (NSFileHandlingPanelOKButton ==
        [oPanel runModalForDirectory: nil
                                file: nil
                               types: [SquishController extensions]])
    {
        [self application: NSApp openFiles: [oPanel filenames]];
    }
}

- (int)showExportOptions
{
    if (nil == exportController)
        exportController =
            [[SquishExportOptionsController alloc] init];

    int result = [NSApp runModalForWindow: [exportController window]];
    [[exportController window] orderOut: self];
    return result;
}


- (NSApplicationDelegateReply)exportFiles: (NSArray *)aFilenames
{
    NSFileManager *fm = [NSFileManager defaultManager];
    NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];
    NSEnumerator *selectionEnum = [aFilenames objectEnumerator];
    NSDirectoryEnumerator *dirEnum;
    NSString *path;
    NSString *file;
    BOOL isDirectory;
    NSSet *extSet = [NSSet setWithArray: [SquishController extensions]];
    int flags = [self flagsFromPrefs];
    int fileMode = [defaults integerForKey: @"fileMode"];
    int mipmapFormat = 0;
    BOOL assembleMipmaps = NO;
    NSString *outputType = nil;
    NSString *dest = [defaults stringForKey: @"browseDest"];
    NSString *destSuffix = @"";
    NSString *destExt = DDS_EXT;

    switch (fileMode)
    {
        case kFileModeDDS:
            switch ([defaults integerForKey: @"outputFormat"])
            {
            case kDxt3:
                outputType = @"DXT3"; break;
            case kDxt5:
                outputType = @"DXT5"; break;
            case kDxt5nm:
                outputType = @"DXT5"; destExt = DXT5NM_EXT; break;
            default:
                outputType = @"DXT1";
            }
            assembleMipmaps   = [defaults boolForKey: @"assembleMipmaps"];
            break;
        case kFileModeMipmap:
            outputType = @"mipmaps";
            mipmapFormat = [defaults integerForKey: @"mipmapFormat"];
            if (mipmapFormat < 0 || mipmapFormat > kMipmapTIFF)
                mipmapFormat = kMipmapTIFF;
            destExt = MIPMAP_EXT[mipmapFormat];
            break;
        default:
            break;
    }
    NSMutableArray *folderContents = [NSMutableArray array];

    while ((path = [selectionEnum nextObject]))
    {
        if ([fm fileExistsAtPath: path isDirectory: &isDirectory])
        {
            if (isDirectory)
            {
                dirEnum = [fm enumeratorAtPath: path];
                while ((file = [dirEnum nextObject]))
                {
                    if ([extSet containsObject: [file pathExtension]])
                    {
                        [folderContents addObject: [path stringByAppendingPathComponent: file]];
                    }
                }
            }
            else
            {
                [folderContents addObject: path];
            }
        }
    }
#ifdef USE_DDS_PREVIEW
    if (nil == ddsPreview)
        ddsPreview = [[SquishDDSWindow alloc] init];
#endif

    int theReplaceChoice = kSquishNoReplace;
    progressController = [[SquishProgressController alloc] init];
    [progressController window];
    [progressController showWindow: self];
    [progressController setProgressIndeterminate: NO];
    [progressController setProgressItemCount: [folderContents count]];
    [progressController setProgressTitleText: [NSString stringWithFormat: NSLocalizedString(@"Exporting to %@...",""), outputType]];
    [progressController setProgressItemText: @""];
    [progressController startProgress];

    NSEnumerator *contentEnum;
    NSAutoreleasePool *pool;
    NSString *outFile;
    NSDate *curTime = [[NSDate alloc] init];
    NSApplicationDelegateReply exitStatus = NSApplicationDelegateReplySuccess;
    int width, height;
    BOOL mipmapsPossible = NO;
    NSBitmapImageRep *rep = nil;

    contentEnum = [folderContents objectEnumerator];

    if (assembleMipmaps)
    {
        NSMutableDictionary *tileDict = [NSMutableDictionary dictionaryWithCapacity: [folderContents count]];
        NSMutableArray *sortedPaths = nil;
        NSArray *sortedTileKeys = nil;
        NSString *tileKey;

        while ((path = [contentEnum nextObject]))
        {
            NSAutoreleasePool *sizeScanPool = [[NSAutoreleasePool alloc] init];
            rep = [NSBitmapImageRep imageRepWithData: [NSData dataWithContentsOfFile: path]];

            if (nil==rep)
            {
                [sizeScanPool release];
                exitStatus = NSApplicationDelegateReplyFailure;
                break;
            }

            tileKey = [NSString stringWithFormat:@"%010d %010d", [rep pixelsWide], [rep pixelsHigh]];
            // Duplicates will be silently ignored
            [tileDict setObject:path forKey:tileKey];
            [sizeScanPool release];
        }

        if (NSApplicationDelegateReplySuccess == exitStatus)
        {
            sortedPaths = [NSMutableArray arrayWithCapacity: [folderContents count]];
            sortedTileKeys = [[tileDict allKeys] sortedArrayUsingSelector:@selector(compare:)];
            NSEnumerator *tileEnum = [sortedTileKeys reverseObjectEnumerator];
            NSArray *dimensions;
            tileKey = [tileEnum nextObject];
            dimensions = [tileKey componentsSeparatedByString: @" "];
            width  = [[dimensions objectAtIndex: 0] intValue];
            height = [[dimensions objectAtIndex: 1] intValue];
            [sortedPaths addObject: [tileDict objectForKey: tileKey]];

            while ((tileKey = [tileEnum nextObject]))
            {
                if (width  > 1) { width  >>= 1; }
                if (height > 1) { height >>= 1; }
                dimensions = [tileKey componentsSeparatedByString: @" "];
                if (width  != [[dimensions objectAtIndex: 0] intValue] ||
                    height != [[dimensions objectAtIndex: 1] intValue])
                {
                    exitStatus = NSApplicationDelegateReplyFailure;
                    break;
                }
                [sortedPaths addObject: [tileDict objectForKey: tileKey]];
            }
            if (width > 1 || height > 1)
                exitStatus = NSApplicationDelegateReplyFailure;
            if (NSApplicationDelegateReplySuccess == exitStatus)
            {
                path = [sortedPaths objectAtIndex: 0];
                [progressController setProgressItemText: [path lastPathComponent]];
                [progressController setProgressItemCount: [sortedPaths count]];
                exitStatus = NSApplicationDelegateReplyCancel;
                outFile = [self pathForOutputFile:path extension:destExt outputFolder:dest];
                outFile = [NSString stringWithFormat:outFile, @""];
                exitStatus = [self confirmReplace:outFile refreshTime:&curTime]; 
                if (NSApplicationDelegateReplySuccess == exitStatus)
                {
                    [[NSDocumentController sharedDocumentController] noteNewRecentDocumentURL: [NSURL fileURLWithPath: path]];
                    exitStatus = [self exportDDSFromFiles:sortedPaths representation:nil toFile:outFile ofType:outputType withFlags:flags];
                }
            }
        }
    }
    else
    {
        while (NSApplicationDelegateReplySuccess == exitStatus &&
               (path = [contentEnum nextObject]))
        {
            [progressController setProgressItemText: [path lastPathComponent]];
            if (![progressController incrementProgress])
                break;
            
            pool = [[NSAutoreleasePool alloc] init];
            mipmapsPossible = NO;
            destSuffix = @"";

            rep = [NSBitmapImageRep imageRepWithData: [NSData dataWithContentsOfFile: path]];
            if (nil == rep)
            {
                [pool release];
                exitStatus = NSApplicationDelegateReplyFailure;
                break;
            }

            [[NSDocumentController sharedDocumentController] noteNewRecentDocumentURL: [NSURL fileURLWithPath: path]];
            outFile = [self pathForOutputFile: path
                                    extension: destExt
                                 outputFolder: dest];
            width  = [rep pixelsWide];
            height = [rep pixelsHigh];
            mipmapsPossible = (width > 0 && height > 0) && (width % 2 == 0) && (height % 2 == 0);
            if (kFileModeMipmap == fileMode)
            {
                if (!mipmapsPossible)
                {
                    [pool release];
                    exitStatus = NSApplicationDelegateReplyFailure;
                    break;
                }
                exitStatus = [self exportMipmapsFromFile:path representation:rep toFile:outFile format:mipmapFormat];
            }
            else
            {
                outFile = [NSString stringWithFormat:outFile, @""];
                if ([progressController incrementProgress])
                    exitStatus = [self confirmReplace:outFile refreshTime:&curTime];
                if (NSApplicationDelegateReplySuccess == exitStatus)
                {
                    exitStatus = [self exportDDSFromFiles:[NSArray arrayWithObject: path] representation:rep toFile:outFile ofType:outputType withFlags:flags];
                }
            }

            [pool release];
        }
    }

    [progressController endProgress];
    theReplaceChoice = [progressController replaceChoice];
    [progressController close];
    [progressController release];
    progressController = nil;
    [curTime release];

    if (NSApplicationDelegateReplyFailure == exitStatus)
    {
        NSRunInformationalAlertPanel(NSLocalizedString(@"Bad stuff happened",""), [NSString stringWithFormat: NSLocalizedString(@"An error occured while processing%@.",""), (path ? [NSString stringWithFormat:@" \"%@\"", path] : @"")], nil, nil, nil);
    }

    return exitStatus;
}


- (NSApplicationDelegateReply)confirmReplace: (NSString *)aOutFile
                                 refreshTime: (NSDate **)aTime
{
    BOOL exists, isFolder;
    NSDate *curTime = nil;
    int theReplaceChoice = kSquishNoReplace;
    NSApplicationDelegateReply exitStatus = NSApplicationDelegateReplySuccess;

    exists = [[NSFileManager defaultManager] fileExistsAtPath: aOutFile isDirectory: &isFolder];
    if (exists && isFolder)
    {
        return NSApplicationDelegateReplyFailure;
    }

    if (exists)
    {
        [progressController showReplaceConfirmIfNecessaryForFile: [aOutFile lastPathComponent]];
        [progressController refresh];
        theReplaceChoice = [progressController replaceChoice];
        
        switch (theReplaceChoice)
        {
            case kSquishStop:
                return NSApplicationDelegateReplyCancel;
            case kSquishNoReplace:
                return NSApplicationDelegateReplySuccess;
            default:
                break;
        }
    }

    curTime = [[NSDate alloc] init];
    if ([curTime timeIntervalSinceDate: *aTime] > 1.0)
    {
        [*aTime release];
        *aTime = curTime;
        if (![progressController refresh])
        {
            return NSApplicationDelegateReplyCancel;
        }
    }
    else
    {
        [curTime release];
    }

    return exitStatus;
}

- (NSApplicationDelegateReply)exportDDSFromFiles: (NSArray *)files
                                  representation: (NSBitmapImageRep *)rep
                                          toFile: (NSString *)aOutFile
                                          ofType: (NSString *)aType
                                       withFlags: (int)aFlags
{
    NSApplicationDelegateReply result = NSApplicationDelegateReplyFailure;
    NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];
    BOOL genMipmapsSetting = [defaults boolForKey:    @"genMipmaps"];
    int  premulOption = 0; //[defaults integerForKey: @"premulAlpha"];
    int profileChoice = kSquishProfileUseEmbedded;

    NSData *profileData;
    CMError matchResult = noErr;
    CMBitmap cmBitmap;

    NSEnumerator *fileEnum = [files objectEnumerator];
    NSString *file;
    FILE *ofp = nil;
    BOOL isMipTile = NO;

    while ((file = [fileEnum nextObject]))
    {
        if (![progressController incrementProgress])
        {
            result = NSApplicationDelegateReplyCancel;
            break;
        }

        if (nil == rep)
            rep = [NSBitmapImageRep imageRepWithData: [NSData dataWithContentsOfFile: file]];
        if (nil == rep)
            break;
        [rep retain];

        memset(&cmBitmap, 0, sizeof(cmBitmap));
        cmBitmap.width      = [rep pixelsWide];
        cmBitmap.height     = [rep pixelsHigh];
        cmBitmap.rowBytes   = [rep bytesPerRow];
        cmBitmap.pixelSize  = [rep bitsPerPixel];
        int samplesPerPixel = [rep samplesPerPixel];
        BOOL hasAlpha = [rep hasAlpha];

        profileData = [rep valueForProperty: NSImageColorSyncProfileData];
        if (profileData)
        {
            CMProfileLocation profileLocation;
            CMProfileRef embedProf;
            profileLocation.locType = cmPtrBasedProfile;
            profileLocation.u.ptrLoc.p = (Ptr)[profileData bytes];
            CMOpenProfile(&embedProf, &profileLocation);
            NSString *embedProfName = (NSString *)CopyProfileDescriptionCFString(embedProf);
//            NSLog(@"Embedded profile = '%@'", embedProfName);
            [embedProfName autorelease];

            [progressController showProfileConfirmIfNecessaryForFile: [file lastPathComponent] embeddedProfileName: embedProfName];
            [progressController refresh];
            profileChoice = [progressController profileChoice];
            switch (profileChoice)
            {
                case kSquishStop:
                    CMCloseProfile(embedProf);
                    [rep release];
                    return NSApplicationDelegateReplyCancel;
                default:
                    break;
            }

            if (hasAlpha)
            {
                cmBitmap.space = cmRGBA32Space;
            }
            else
            {
                int bps = [rep bitsPerSample];
                int bpp = [rep bitsPerPixel];
                switch (bps)
                {
                    case 5:
                        cmBitmap.space = cmRGB16Space;
                        break;
                    case 8:
                        cmBitmap.space = (bpp == 24) ? cmRGB24Space : cmRGB32Space;
                        break;
                    default:
                        cmBitmap.space = cmRGBA32Space;
                        break;
                }
            }
            BOOL colorMatchFailed = NO;
            CMWorldRef cw = nil;
            cmBitmap.image = (char *)[rep bitmapData];

            if (kSquishProfileOverride == profileChoice)
            {
                switch ([progressController profileOverrideChoice])
                {
                    case 0:
                    {
                        // Ignore embedded profile
                        break;
                    }
                    case cmSRGBData:
                    {
                        CMProfileRef srgbProf = CreateICCsRGBProfile();
                        if (srgbProf && dispProf)
                        {
                            NCWNewColorWorld(&cw, srgbProf, dispProf);
                            matchResult = CWMatchBitmap(cw, &cmBitmap, NULL, NULL, NULL);
                            CMCloseProfile(srgbProf);
                        }
                        else
                        {
                            colorMatchFailed = YES;
                        }
                        break;
                    }
                    default:
                        colorMatchFailed = YES;
                }
            }
            else if (embedProf)
            {
                if (dispProf)
                {
                    NCWNewColorWorld(&cw, embedProf, dispProf);
                    matchResult = CWMatchBitmap(cw, &cmBitmap, NULL, NULL, NULL);
                }
                else
                {
                    colorMatchFailed = YES;
                }
            }
            
            if (cw)
                CWDisposeColorWorld(cw);
            if (embedProf)
                CMCloseProfile(embedProf);

            if (colorMatchFailed || noErr != matchResult)
            {
                return result;
            }
        }

        BOOL genMipmaps = (genMipmapsSetting &&
                           (cmBitmap.width % 2 == 0) && (cmBitmap.height % 2 == 0));
        if (isMipTile && !genMipmaps)
        {
            [rep release];
            break;
        }

        if (!isMipTile)
        {
            ofp = fopen([aOutFile UTF8String], "wb");
            if (ofp)
            {
                result = [self writeDDSHeaderToFile: ofp
                                             ofType: aType
                                              width: cmBitmap.width
                                             height: cmBitmap.height
                                          withFlags: aFlags
                                            mipmaps: genMipmaps];
                if (NSApplicationDelegateReplySuccess != result)
                {
                    fclose(ofp);
                    [rep release];
                    return result;
                }
            }
            if (NSApplicationDelegateReplySuccess != result)
            {
                [rep release];
                return result;
            }
        }

        result = [self compressBytes: [rep bitmapData]
                              toFile: ofp
                               width: cmBitmap.width
                              height: cmBitmap.height
                         bytesPerRow: cmBitmap.rowBytes
                     samplesPerPixel: samplesPerPixel
                               alpha: hasAlpha
                         premultiply: premulOption
                           withFlags: aFlags];

        if (NSApplicationDelegateReplySuccess == result &&
            genMipmaps &&
            [files count] < 2)
        {
            NSBitmapImageRep *miprep = nil;
            while (cmBitmap.width > 1 || cmBitmap.height > 1)
            {
                if (cmBitmap.width  > 1) { cmBitmap.width  >>= 1; }
                if (cmBitmap.height > 1) { cmBitmap.height >>= 1; }
                
                miprep = [SquishController mipmapFromRepresentation:rep width:cmBitmap.width height:cmBitmap.height];
                [miprep retain];

                result = [self compressBytes: [miprep bitmapData]
                                      toFile: ofp
                                       width: cmBitmap.width
                                      height: cmBitmap.height
                                 bytesPerRow: [miprep bytesPerRow]
                             samplesPerPixel: [miprep samplesPerPixel]
                                       alpha: [miprep hasAlpha]
                                 premultiply: premulOption
                                   withFlags: aFlags];

                [rep release];
                rep = miprep;

                if (NSApplicationDelegateReplySuccess != result)
                    break;
            }
        }
        [rep release]; rep = nil;
        isMipTile = YES;    // subsequent files are mipmap tiles
    }

    if (ofp) fclose(ofp);
    return result;
}

- (NSApplicationDelegateReply)exportMipmapsFromFile: (NSString *)path
                                     representation: (NSBitmapImageRep *)rep
                                             toFile: (NSString *)aOutFile
                                             format: (int)mipmapFormat
{
    NSApplicationDelegateReply result = NSApplicationDelegateReplyCancel;
    int width  = [rep pixelsWide];
    int height = [rep pixelsHigh];
    NSString *outFile;
    NSDate *curTime = [[NSDate alloc] init];

    outFile = [NSString stringWithFormat:aOutFile, [NSString stringWithFormat: @"_%dx%d", width, height]];
    if ([progressController incrementProgress])
        result = [self confirmReplace:outFile refreshTime:&curTime];
    if (NSApplicationDelegateReplySuccess != result)
    {
        [curTime release];
        return result;
    }

    NSBitmapImageFileType exportType;
    NSMutableDictionary *exportProps = [[NSMutableDictionary alloc] init];
    id srcProp;

    if ((srcProp = [rep valueForProperty: NSImageColorSyncProfileData]))
        [exportProps setObject:srcProp forKey: NSImageColorSyncProfileData];
    switch (mipmapFormat)
    {
        case kMipmapPNG:
            exportType = NSPNGFileType;
            if ((srcProp = [rep valueForProperty: NSImageInterlaced]))
                [exportProps setObject:srcProp forKey:NSImageInterlaced];
            if ((srcProp = [rep valueForProperty: NSImageGamma]))
                [exportProps setObject:srcProp forKey:NSImageGamma];
            break;
        case kMipmapTIFF:
        default:
            exportType = NSTIFFFileType;
            if ((srcProp = [rep valueForProperty: NSImageCompressionMethod]))
                [exportProps setObject:srcProp forKey:NSImageCompressionMethod];
            else
                [exportProps setObject:[NSNumber numberWithInt:NSTIFFCompressionLZW] forKey:NSImageCompressionMethod];
            break;
    }

    NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];

    NSData *exportData = [rep representationUsingType:exportType properties:exportProps];
    if (exportData)
    {
        if (![exportData writeToFile:outFile atomically:NO])
            result = NSApplicationDelegateReplyFailure;
    }
    else
    {
        result = NSApplicationDelegateReplyFailure;
    }

    [rep retain];
    NSBitmapImageRep *miprep = nil;

    [pool release];

    while (width > 1 || height > 1)
    {
        pool = [[NSAutoreleasePool alloc] init];
        if (width  > 1) { width  >>= 1; }
        if (height > 1) { height >>= 1; }
        outFile = [NSString stringWithFormat:aOutFile, [NSString stringWithFormat: @"_%dx%d", width, height]];
        result = NSApplicationDelegateReplyCancel;
        if ([progressController incrementProgress])
            result = [self confirmReplace:outFile refreshTime:&curTime];
        if (NSApplicationDelegateReplySuccess != result)
            break;
        miprep = [SquishController mipmapFromRepresentation:rep width:width height:height];
        [miprep retain];
        srcProp = [exportProps objectForKey: NSImageColorSyncProfileData];
        if (srcProp)
            [miprep setProperty:NSImageColorSyncProfileData withValue:srcProp];
        exportData = nil;

        switch (mipmapFormat)
        {
            case kMipmapPNG:
            {
                exportData = nil;
#ifdef USE_LIBPNG
                png_structp png_ptr;
                png_infop   info_ptr;
                png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, nil, nil, nil);
                info_ptr = png_create_info_struct(png_ptr);
                if (!info_ptr)
                {
                    png_destroy_write_struct(&png_ptr, (png_infopp)nil);
                    break;
                }
                if (setjmp(png_jmpbuf(png_ptr)))
                {
                    png_destroy_write_struct(&png_ptr, &info_ptr);
                    break;
                }
                png_set_write_fn(png_ptr, <#png_voidp io_ptr#>, PngWrite, PngFlush);
                png_set_compression_level(png_ptr, Z_DEFAULT_COMPRESSION);
                int pngColorType = PNG_COLOR_TYPE_RGB_ALPHA;
                srcProp = [exportProps objectForKey: NSImageInterlaced];
                png_set_IHDR(png_ptr, info_ptr, width, height, [miprep bitsPerPixel], <#int color_type#>, ([srcProp boolValue] ? PNG_INTERLACE_ADAM7 : PNG_INTERLACE_NONE), PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
                srcProp = [exportProps objectForKey: NSImageGamma];
                if (srcProp)
                {
                    png_set_gAMA(png_ptr, info_ptr, [srcProp doubleValue]);
                }
                srcProp = [exportProps objectForKey: NSImageColorSyncProfileData];
                if (srcProp)
                {
                    png_set_iCCP(png_ptr, info_ptr, <#png_charp name#>, <#int compression_type#>, <#png_charp profile#>, <#png_uint_32 proflen#>);
                }
                png_write_png(png_ptr, info_ptr, PNG_TRANSFORM_IDENTITY, nil);
                png_destroy_write_struct(&png_ptr, &info_ptr);
#endif
                break;
            }
            case kMipmapTIFF:
            default:
                srcProp = [exportProps objectForKey: NSImageCompressionMethod];
                if (srcProp)
                    [miprep setCompression:(NSTIFFCompression)[srcProp intValue] factor:0.5];
                exportData = [miprep TIFFRepresentation];
                break;
        }
        if (exportData)
        {
            if (![exportData writeToFile:outFile atomically:NO])
                result = NSApplicationDelegateReplyFailure;
        }
        [rep release];
        rep = miprep;
        [pool release];
    }

    [rep release];
    [exportProps release];
    [curTime release];

    return result;
}

+ (NSBitmapImageRep *)mipmapFromRepresentation: (NSBitmapImageRep *)rep
                                         width: (int)width
                                        height: (int)height
{
    unsigned char *row = [rep bitmapData];
    int rowbytes  = [rep bytesPerRow];
    
    NSBitmapImageRep *miprep = [[NSBitmapImageRep alloc] initWithBitmapDataPlanes:nil pixelsWide:width pixelsHigh:height bitsPerSample:[rep bitsPerSample] samplesPerPixel:[rep samplesPerPixel] hasAlpha:[rep hasAlpha] isPlanar:[rep isPlanar] colorSpaceName:[rep colorSpaceName] bytesPerRow:0 bitsPerPixel:[rep bitsPerPixel]];
    
    unsigned char *scaled = [miprep bitmapData];
    int scaledRowBytes    = [miprep bytesPerRow];
    int samplesPerPixel   = [miprep samplesPerPixel];
    int samplesPerRow     = width*samplesPerPixel;
    int x, y, c, k;
    int *scaledBuf = (int *)calloc(1, sizeof(int)*samplesPerRow);
    for (y = 0; y < height; ++y)
    {
        for (x = 0; x < samplesPerRow; x+=samplesPerPixel)
        {
            for (c = 0; c < samplesPerPixel; ++c)
            {
                k = 2*x+c;
                scaledBuf[x+c] = row[k] + row[k+samplesPerPixel];
            }
        }
        
        row += rowbytes;
        for (x = 0; x < samplesPerRow; x+=samplesPerPixel)
        {
            for (c = 0; c < samplesPerPixel; ++c)
            {
                k = 2*x+c;
                scaled[x+c] = (unsigned char)FLOOR(0.25*(scaledBuf[x+c]+row[k]+row[k+samplesPerPixel])+0.5);
            }
        }
        row    += rowbytes;
        scaled += scaledRowBytes;
    }
    free(scaledBuf);
    return [miprep autorelease];
}


- (NSApplicationDelegateReply)compressBytes: (unsigned char *)aBytes
                                    toBlock: (u8 *)output
                                      width: (size_t)aWidth
                                     height: (size_t)aHeight
                                bytesPerRow: (size_t)aBytesPerRow
                            samplesPerPixel: (int)aSamplesPerPixel
                                      alpha: (int)aAlpha
                                  withFlags: (int)aFlags
{
    NSApplicationDelegateReply result = NSApplicationDelegateReplyFailure;
    if (NULL == output) return result;
    SquishCompressNotifier progressNotifier(progressController);

    CompressImage(aBytes,
                  aWidth, aHeight,
                  aBytesPerRow, aSamplesPerPixel,
                  aAlpha,
                  output,
                  aFlags,
                  &progressNotifier);
    
    if (progressNotifier.isAborted())
    {
        return NSApplicationDelegateReplyCancel;
    }
    return NSApplicationDelegateReplySuccess;
}

- (NSApplicationDelegateReply)writeDDSHeaderToFile: (FILE *)ofp
                                            ofType: (NSString *)aType
                                             width: (size_t)aWidth
                                            height: (size_t)aHeight
                                         withFlags: (int)aFlags
                                           mipmaps: (BOOL)genMipmaps
{
    NSApplicationDelegateReply result = NSApplicationDelegateReplyFailure;
    if (NULL == ofp) return result;

    int byteSize = GetStorageRequirements(aWidth, aHeight, aFlags);
    uint32_t mipLevels = 1;
    uint32_t ddsFlags = DDSD_CAPS | DDSD_PIXELFORMAT | DDSD_WIDTH | DDSD_HEIGHT | DDSD_LINEARSIZE;
    uint32_t ddsCaps = DDSCAPS_TEXTURE;
    
    if (genMipmaps)
    {
        size_t mipw = aWidth;
        size_t miph = aHeight;
        while (mipw > 1 || miph > 1)
        {
            if (mipw > 1) mipw >>= 1;
            if (miph > 1) miph >>= 1;
            byteSize += GetStorageRequirements(mipw, miph, aFlags);
            ++mipLevels;
        }
        if (mipLevels > 1)
        {
            ddsFlags |= DDSD_MIPMAPCOUNT;
            ddsCaps  |= DDSCAPS_COMPLEX | DDSCAPS_MIPMAP;
        }
    }

    char ddsMagic[4];
    ddsMagic[0] = 'D';
    ddsMagic[1] = 'D';
    ddsMagic[2] = 'S';
    ddsMagic[3] = ' ';
    size_t ddsmsize = sizeof ddsMagic;
    fwrite(ddsMagic, ddsmsize, 1, ofp);
    
    DDSurfaceDesc ddsd = {0};
    ddsd.size = CFSwapInt32HostToLittle(sizeof ddsd);
    ddsd.flags = CFSwapInt32HostToLittle(ddsFlags);
    ddsd.width = CFSwapInt32HostToLittle(aWidth);
    ddsd.height = CFSwapInt32HostToLittle(aHeight);
    ddsd.pitch = CFSwapInt32HostToLittle(byteSize);
    ddsd.mipMapLevels = CFSwapInt32HostToLittle(mipLevels);
    ddsd.format.size = CFSwapInt32HostToLittle(sizeof ddsd.format);
    ddsd.format.flags = CFSwapInt32HostToLittle(DDPF_FOURCC);
    ddsd.format.fourCC = CFSwapInt32HostToLittle(FourCC([aType UTF8String]));
    ddsd.caps.caps = CFSwapInt32HostToLittle(ddsCaps);
    
    fwrite(&ddsd, sizeof ddsd, 1, ofp);

    return NSApplicationDelegateReplySuccess;
}

- (NSApplicationDelegateReply)compressBytes: (unsigned char *)aBytes
                                     toFile: (FILE *)ofp
                                      width: (size_t)aWidth
                                     height: (size_t)aHeight
                                bytesPerRow: (size_t)aBytesPerRow
                            samplesPerPixel: (int)aSamplesPerPixel
                                      alpha: (int)aAlpha
                                premultiply:
#ifdef USE_AUTO_PREMULTIPLY
                                             (BOOL)aPremultiply
#else
                                             (int)aPremulOption
#endif
                                  withFlags: (int)aFlags
{
    NSApplicationDelegateReply result = NSApplicationDelegateReplyFailure;
    int byteSize = GetStorageRequirements(aWidth, aHeight, aFlags);
    u8 *output = (u8 *)calloc(1, byteSize);
    if (NULL == output) return result;

    int format = [[NSUserDefaults standardUserDefaults] integerForKey: @"outputFormat"];
    if (kDxt5nm == format)
    {
#ifdef USE_AUTO_PREMULTIPLY
        aPremultiply = NO;
#else
        aPremulOption = kAlphaUnchanged;
#endif
        ArrangeNormalMapChannels(aBytes, aWidth, aHeight, aSamplesPerPixel);
    }

#ifdef USE_AUTO_PREMULTIPLY
    if (aPremultiply)
    {
        PremultiplyImageAlpha(aBytes, aWidth, aHeight, aBytesPerRow);
    }
#else
    switch (aPremulOption)
    {
        case kAlphaPremultiply:
            PremultiplyImageAlpha(aBytes, aWidth, aHeight, aBytesPerRow);
            break;
        case kAlphaUnpremultiply:
            UnpremultiplyImageAlpha(aBytes, aWidth, aHeight, aBytesPerRow);
            break;
        default:
            break;
    }
#endif

    result = [self compressBytes: aBytes
                         toBlock: output
                           width: aWidth
                          height: aHeight
                     bytesPerRow: aBytesPerRow
                 samplesPerPixel: aSamplesPerPixel
                           alpha: aAlpha
                       withFlags: aFlags];
    
    if (NSApplicationDelegateReplySuccess == result)
    {
        fwrite(output, byteSize, 1, ofp);
#if 0
        BOOL doPreview = NO;
        if (doPreview)
        {
        u8 *decomp = NULL;
        result = [self decompressBytes: output
                               toBlock: &decomp
                                 width: aWidth
                                height: aHeight
                             withFlags: aFlags];
        if (NSApplicationDelegateReplySuccess == result &&
            aWidth > 64)
        {
            NSBitmapImageRep *decompRep = [[NSBitmapImageRep alloc] initWithBitmapDataPlanes:&output pixelsWide:aWidth pixelsHigh:aHeight bitsPerSample:8 samplesPerPixel:aSamplesPerPixel hasAlpha:YES isPlanar:NO colorSpaceName:NSCalibratedRGBColorSpace bytesPerRow:aBytesPerRow bitsPerPixel:32];
            if (decompRep)
            {
                NSData *previewData = [decompRep TIFFRepresentation];
                [previewData writeToFile:@"preview.tiff" atomically:NO];
#ifdef USE_DDS_PREVIEW
                unsigned char *decompDestData = [decompRep bitmapData];
                memcpy(decompDestData, decomp, aBytesPerRow * aHeight);
                NSImage *decompImg = [[NSImage alloc] initWithSize: NSMakeSize(aWidth, aHeight)];
                if (decompImg)
                {
                    [decompImg addRepresentation: [decompRep autorelease]];
                    [decompImg setScalesWhenResized: YES];
                    [decompImg setSize: NSMakeSize(aWidth, aHeight)];
                    [ddsPreview setAfterImage: [decompImg autorelease]];
                }
#endif
            }
        }
        if (decomp)
            free(decomp);
        }
#endif
    }

    free(output);
    return result;
}

- (NSApplicationDelegateReply)decompressBytes: (u8 *)aBytes
                                      toBlock: (u8 **)output
                                        width: (size_t)aWidth
                                       height: (size_t)aHeight
                                    withFlags: (int)aFlags
{
    NSApplicationDelegateReply result = NSApplicationDelegateReplyFailure;
    if (NULL == output) return result;
    *output = (u8 *)calloc(1, 4 * aWidth * aHeight);
    DecompressImage(*output, aWidth, aHeight, aBytes, aFlags);
    result = NSApplicationDelegateReplySuccess;
    return result;
}

+ (NSArray *)extensions
{
    return [NSArray arrayWithObjects:
            @"bmp", @"gif", @"jpg", @"jpeg", @"jpe", @"png", @"tiff", @"tif",
            nil];
}

- (int)flagsFromPrefs
{
    NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];
    int flags = 0;
    int format = [defaults integerForKey: @"outputFormat"];
    flags  = (kDxt5nm == format) ? kDxt5 : format;
    flags |= [defaults integerForKey: @"quality"];
    if ([defaults boolForKey: @"weightByAlpha"])
        flags |= kWeightColourByAlpha;
    flags |= [defaults integerForKey: @"metric"];

    return flags;
}

- (NSString *)pathForOutputFile: (NSString *)aFile
                      extension: (NSString *)aExt
                   outputFolder: (NSString *)aDest
{
    NSString *result = [[[aFile stringByDeletingPathExtension] stringByAppendingString:@"%@"] stringByAppendingPathExtension: aExt];
    if (aDest)
    {
        result = [aDest stringByAppendingPathComponent: [result lastPathComponent]];
    }
    return result;
}


- (void)browseDestination: (id)aSender
{
    NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];
    NSString *selection = nil;
    NSOpenPanel *oPanel = [NSOpenPanel openPanel];
    [oPanel setAllowsMultipleSelection: NO];
    [oPanel setCanChooseDirectories: YES];
    [oPanel setCanChooseFiles: NO];
    [oPanel setCanCreateDirectories: YES];

    if (NSFileHandlingPanelOKButton ==
        [oPanel runModalForDirectory:nil file:nil])
    {
        NSArray *filenames = [oPanel filenames];
        if (filenames && [filenames count] > 0)
            selection = [filenames objectAtIndex: 0];
    }

    [defaults setObject: selection forKey: @"browseDest"];
}

- (void)convert: (id)aSender
{
    [NSApp stopModal];
}

- (void)abortConvert: (id)aSender
{
    [NSApp abortModal];
}

- (void)showMyHelp: (id)aSender
{
    [helpWindow setWorksWhenModal: YES];
    [helpWindow makeKeyAndOrderFront: aSender];
}

- (BOOL)enableGenMipmaps
{
    return enableGenMipmaps;
}

- (BOOL)enableAssembleMipmaps
{
    return enableAssembleMipmaps;
}
- (void)setEnableGenMipmaps: (BOOL)aEnable { enableGenMipmaps = aEnable; }
- (void)setEnableAssembleMipmaps: (BOOL)aEnable { enableAssembleMipmaps = aEnable; }
@end
