/*
 *  imageutils.c
 *  squish
 *
 *  Created by Da Woon Jung on 2007-05-04.
 *  Copyright 2007 Da Woon Jung. All rights reserved.
 */

#include "imageutils.h"


/* Colorsync Code taken from Apple Technical Q&A QA1205, QA1396 */

CMProfileRef CreateICCProfileFromPath(const char *iccProfilePath)
{
    CMProfileRef    iccProfile = (CMProfileRef)NULL;
    CMProfileLocation loc;
    
    /* Specify that the location of the profile will be a POSIX path to the profile. */
    loc.locType = cmPathBasedProfile;
    
    /* Make sure the path is not larger then the buffer */
    if(strlen(iccProfilePath) > sizeof(loc.u.pathLoc.path))
        return NULL;
    
    /* Copy the path the profile into the CMProfileLocation structure */
    strcpy(loc.u.pathLoc.path, iccProfilePath);
    
    /* Open the profile */
    if (CMOpenProfile(&iccProfile, &loc) != noErr)
    {
        iccProfile = (CMProfileRef)NULL;
        return NULL;
    }
    
    return iccProfile;
}

CMProfileRef CreateProfileFromSystemICCProfileName(CFStringRef profileName)
{
    FSRef pathToProfilesFolder;
    FSRef pathToProfile;
    
    /* Find the Systems Color Sync Profiles folder */
    if(FSFindFolder(kOnSystemDisk, kColorSyncProfilesFolderType,
                    kDontCreateFolder, &pathToProfilesFolder) == noErr)
    {
        /* Make a UniChar string of the profile name */
        UniChar uniBuffer[sizeof(CMPathLocation)];
        CFStringGetCharacters (profileName,CFRangeMake(0,CFStringGetLength(profileName)),uniBuffer);
        
        /* Create a FSRef to the profile in the Systems Color Sync Profile folder */
        if(FSMakeFSRefUnicode (&pathToProfilesFolder,CFStringGetLength(profileName),uniBuffer,
                               kUnicodeUTF8Format,&pathToProfile) == noErr)
        {
            UInt8 path[sizeof(CMPathLocation)];
            
            /* Write the posix path to the profile into our path buffer from the FSRef */
            if(FSRefMakePath(&pathToProfile, path, sizeof(CMPathLocation)) == noErr)
                return CreateICCProfileFromPath((const char *)path);
        }
    }
    
    return NULL;
}

CMProfileRef CreateICCGenericRGBProfile()
{
    return CreateProfileFromSystemICCProfileName(CFSTR("Generic RGB Profile.icc"));
}

CMProfileRef CreateICCGenericGrayProfile()
{
    return CreateProfileFromSystemICCProfileName(CFSTR("Generic Gray Profile.icc"));
}

CMProfileRef CreateICCsRGBProfile()
{
    return CreateProfileFromSystemICCProfileName(CFSTR("sRGB Profile.icc"));
}

CMProfileRef CreateSystemProfile()
{
    CMProfileRef sysprof = NULL;
    
    // Get the System Profile for the main display
    if (CMGetSystemProfile(&sysprof) != noErr)
    {
        return NULL;
    }
    
    return sysprof;
}

CFStringRef CopyProfileDescriptionCFString(CMProfileRef prof)
{
    Str255         profileName;
    ScriptCode     code;
    CFStringRef    str = nil;
    CMError        err;
    
    /* for v4 profiles, try to get the best localized name from the 'desc'/'mluc' tag */
    err = CMCopyProfileLocalizedString(prof, cmProfileDescriptionTag, 0,0, &str);
    /* if that didn't work... */
    if (err != noErr)
    {
        /* for Apple's localized v2 profiles, try to get the best localized name from the 'dscm'/'mluc' tag */
        err = CMCopyProfileLocalizedString(prof, cmProfileDescriptionMLTag, 0,0, &str);
        /* if that didn't work... */
        if (err != noErr)
        {
            /* for normal v2 profiles, get the name from the 'desc'/'desc' tag */
            err = CMGetScriptProfileDescription( prof, profileName, &code);
            /* convert it to a CFString */
            if (err == noErr)
            {
                str = CFStringCreateWithPascalString(0, profileName, code);
            }
        }
    }
    return str;
}


/* Premultiply an image's alpha. Expects ARGB layout. */
void PremultiplyImageAlpha(unsigned char *data,
                           size_t width, size_t height,
                           size_t bytesPerRow)
{
    vImage_Buffer vbuf;
    vbuf.data = data;
    vbuf.width = width;
    vbuf.height = height;
    vbuf.rowBytes = bytesPerRow;
    vImagePremultiplyData_ARGB8888(&vbuf, &vbuf, 0);

//    else if (kCGImageAlphaPremultipliedLast == alphaInfo)
//        vImagePremultiplyData_RGBA8888(&vbuf, &vbuf, 0);
}

/* Unpremultiply an image's alpha. Expects ARGB layout. */
void UnpremultiplyImageAlpha(unsigned char *data,
                             size_t width, size_t height,
                             size_t bytesPerRow)
{
    vImage_Buffer vbuf;
    vbuf.data = data;
    vbuf.width = width;
    vbuf.height = height;
    vbuf.rowBytes = bytesPerRow;
    vImageUnpremultiplyData_ARGB8888(&vbuf, &vbuf, 0);
}

/* Swizzles <a,r,g,b> => <r,0,g,0> */
void ArrangeNormalMapChannels(unsigned char *data,
                              size_t width, size_t height,
                              int samplesPerPixel)
{
    size_t x, y;
    unsigned char *p = data;

    for (y = 0; y < height; ++y)
    {
        for (x = 0; x < width; ++x,
             p += samplesPerPixel)
        {
            p[0] = p[1];
            p[1] = 0;
            p[3] = 0;
        }
    }
}
